"""
The qiprofile imaging Mongodb data model.
"""

import re
import decimal
from numbers import Number
import mongoengine
from mongoengine import (fields, signals, ValidationError)
from .. import choices
from .encounter import Encounter

class Session(Encounter):
    """The MR session (a.k.a. *study* in DICOM terminology)."""

    modelings = fields.ListField(
        field=fields.EmbeddedDocumentField('Modeling')
    )
    """The modeling performed on the session."""

    detail = fields.ReferenceField('SessionDetail')
    """The session detail reference."""

    @classmethod
    def pre_delete(cls, sender, document, **kwargs):
        """Cascade delete this Session's detail."""

        if self.detail:
            self.detail.delete()

signals.pre_delete.connect(Session.pre_delete, sender=Session)


class VoxelSize(mongoengine.EmbeddedDocument):
    """The voxel width, depth and spacing."""
    
    width = fields.FloatField()
    """
    The voxel width (= voxel length) in millimeters. For an MR, the width
    is the DICOM Pixel Spacing value.
    """
    
    depth = fields.FloatField()
    """
    The voxel depth in millimeters. For an MR, the depth is the DICOM
    Slice Thickness.
    """
    
    spacing = fields.FloatField()
    """
    The inter-slice spacing in millimeters. For an MR, the spacing is
    the DICOM Spacing Between Slices.
    """


class Point(mongoengine.EmbeddedDocument):
    """The 3D point in the volume voxel space."""
    
    x = fields.IntField()
    """
    The x dimension value in the image coordinate system.
    """
    
    y = fields.IntField()
    """
    The y dimension value in the image coordinate system.
    """
    
    z = fields.IntField()
    """
    The z dimension value in the image coordinate system.
    """


class Region(mongoengine.EmbeddedDocument):
    """The 3D region in volume voxel space."""

    mask = fields.StringField()
    """The binary mask file name."""

    label_map = fields.EmbeddedDocumentField('LabelMap')
    """The label map file name."""
        
    centroid = mongoengine.EmbeddedDocumentField(Point)
    """The region centroid."""

    average_intensity = fields.FloatField()
    """The average signal in the region."""


class Volume(mongoengine.EmbeddedDocument):
    """The 3D image volume."""

    filename = fields.StringField(required=True)
    """The image file pathname relative to the web app root."""

    average_intensity = fields.FloatField()
    """The image signal intensity over the entire volume."""


class LabelMap(mongoengine.EmbeddedDocument):
    """A label map with an optional associated color lookup table."""

    filename = fields.StringField(required=True)
    """The label map file path relative to the web app root."""

    color_table = fields.StringField()
    """The color map lookup table file path relative to the web app root."""


class ImageSequence(mongoengine.EmbeddedDocument):
    """The scan or registration image volume container."""

    meta = dict(allow_inheritance=True)
    
    volumes = fields.ListField(field=mongoengine.EmbeddedDocumentField(Volume))
    """The 3D volume images in the sequence."""

class RegistrationProtocol(mongoengine.Document):
    """The registration settings."""

    meta = dict(collection='qiprofile_registration_protocol')

    technique = fields.StringField(required=True)
    """The registration technique, e.g. ``ANTS`` or ``FNIRT``."""

    parameters = fields.DictField()
    """The registration input {parameter: value} dictionary."""


class Registration(ImageSequence):
    """
    The patient image registration that results from processing a scan.
    """
    
    protocol = fields.ReferenceField(RegistrationProtocol, required=True)
    """The registration protocol."""

    resource = fields.StringField(required=True)
    """The registration imaging store resource name, e.g. ``reg_k3RtZ``."""


class ScanProtocol(mongoengine.Document):
    """
    The scan acquisition protocol. Scans with the same protocol
    and image dimensions are directly comparable, e.g. in comparing
    modeling results across subjects or sessions.
    """

    meta = dict(collection='qiprofile_scan_protocol')
    
    ORIENTATION = ['axial', 'sagittal', 'coronal']
    """The three imaging axes."""

    scan_type = fields.StringField(required=True)
    """
    The scan type designation controlled value, e.g. ``T1``. The REST update
    client is responsible for ensuring that scan type synonyms resolve to
    the same scan type value, e.g. scans with descriptions including ``t2``,
    ``T2`` and ``T2W`` should all resolve to scan type ``T2``.
    """

    orientation = fields.StringField(choices=ORIENTATION,
                                     max_length=choices.max_length(ORIENTATION))
    """The imaging :const:`ORIENTATION` controlled value."""
    
    technique = fields.StringField()
    """
    The scan technique, e.g. ``STIR``, ``FLAIR`` or ``BLISS``. The REST update
    client is responsible for ensuring that technique synonyms resolve to the
    same technique value, e.g. scans with descriptions including ``BLISS`` and
    ``BLISS_AUTO_SENSE`` should both resolve to technique ``BLISS``.
    """

    description = fields.StringField()
    """
    The image acquisition scan description, e.g. ``T1 AX SPIN ECHO``.
    This field is customarily specified as the DICOM
    *Series Description* or *Protocol Name* tag. This field is often
    used by the REST update client to infer the controlled vocabulary
    scan type, orientation and technique field values.
    """
    
    voxel_size = fields.EmbeddedDocumentField(VoxelSize)
    """The voxel size in millimeters."""


class Scan(ImageSequence):
    """
    The the concrete subclass of the abstract :class:`ImageSequence`
    class for scans.
    """

    number = fields.IntField(required=True)
    """
    The scan number. In the XNAT image store, each scan is
    identified by a number unique within the session.
    """

    protocol = fields.ReferenceField(ScanProtocol, required=True)
    """The scan acquisition protocol."""

    bolus_arrival_index = fields.IntField()
    """The bolus arrival volume index."""
    
    rois = fields.ListField(fields.EmbeddedDocumentField('Region'))
    """The image regions of interest. There is one ROI per scan lesion."""

    registrations = fields.ListField(
        field=fields.EmbeddedDocumentField(Registration)
    )
    """
    The registrations performed on the scan.
    """

    def clean(self):
        """Verify that the bolus arrival index references a volume,."""
        arv = self.bolus_arrival_index
        if arv:
            if not self.volumes:
                raise ValidationError("Session does not have a volume")
            if arv < 0 or arv >= len(self.volumes):
                raise ValidationError(("Bolus arrival index does not refer"
                                       " to a valid volume index: %d") % arv)


class ModelingProtocol(mongoengine.Document):
    """The modeling procedure inputs."""

    meta = dict(collection='qiprofile_modeling_protocol')
    
    technique = fields.StringField(required=True)
    """
    The modeling algorithm or framework, e.g. ``Tofts``.
    """

    input_parameters = fields.DictField()
    """
    The modeling execution input {*parameter*: *value*} dictionary.
    """


class Modeling(mongoengine.EmbeddedDocument):
    """
    The QIN pharmicokinetic modeling run on an image sequence.
    """

    class ParameterResult(mongoengine.EmbeddedDocument):
        """The output for a given modeling run result parameter."""

        filename = fields.StringField(required=True)
        """The voxel-wise mapping file path relative to the web app root."""

        average = fields.FloatField()
        """The average parameter value over all voxels."""

        label_map = fields.EmbeddedDocumentField('LabelMap')
        """The label map overlay NiFTI file."""
    
    class Source(mongoengine.EmbeddedDocument):
        """
        This Modeling.Source embedded class works around the following
        mongoengine limitation:
        
        * mongoengine does not allow heterogeneous collections, i.e.
          a domain model Document subclass can not have subclasses.
          Furthermore, the domain model Document class cannot be
          an inner class, e.g. ModelingProtocol.
        
        Consequently, the Modeling.source field cannot represent an
        abstract superclass of ScanProtocol and RegistrationProtocol.
        This Source embedded document introduces a disambiguation
        level by creating a disjunction object that can either hold
        a *registration* reference or a *scan* reference.
        """
        
        scan = fields.ReferenceField(ScanProtocol)
        
        registration = fields.ReferenceField(RegistrationProtocol)
    
    protocol = fields.ReferenceField(ModelingProtocol, required=True)
    """The modeling protocol."""
    
    source = fields.EmbeddedDocumentField(Source, required=True)
    """The modeling source protocol."""

    resource = fields.StringField(required=True)
    """The modeling imaging store resource name, e.g. ``pk_R3y9``."""

    result = fields.DictField(
        field=mongoengine.EmbeddedDocumentField(ParameterResult)
    )
    """
    The modeling {*parameter*: *result*} dictionary, where:
    
    - *parameter* is the lower-case underscore parameter key, e.g.
      ``k_trans``.
    
    - *result* is the corresponding :class:`ParameterResult`

    The parameters are determined by the :class:`ModelingProtocol`
    technique. For example, the `OHSU QIN modeling workflow`_ includes
    the following outputs for the FXL (`Tofts standard`_) model and the
    FXR (`shutter speed`_) model:
    
    - *fxl_k_trans*, *fxr_k_trans*: the |Ktrans| vascular permeability
       transfer constant

    - *delta_k_trans*: the FXR-FXL |Ktrans| difference

    - *fxl_v_e*, *fxr_v_e*: the |ve| extravascular extracellular volume
       fraction

    - *fxr_tau_i*: the |taui| intracellular |H2O| mean lifetime

    - *fxl_chi_sq*, *fxr_chi_sq*: the |chisq| intensity goodness of fit
    
    The REST client is responsible for anticipating and interpreting the
    meaning of the *parameter* based on the modeling technique. For
    example, if the image store has a session modeling resource
    ``pk_h7Jtl`` which includes the following files::
    
        k_trans.nii.gz
        k_trans_overlay.nii.gz
        chi_sq.nii.gz
        chi_sq_overlay.nii.gz
    
    then a REST database update client might calculate the average |Ktrans|
    and |chisq| values and populate the REST database as follows::
    
        t1 = ScanProtocol.get_or_create(scan_type='T1',
                                        defaults=dict(orientation='axial'))
        tofts = ModelingProtocol.get_or_create(technique='Tofts')
        ktrans_label_map = LabelMap(filename='k_trans_overlay.nii.gz',
                                    color_table='jet.txt')
        ktrans = Modeling.ParameterResult(filename='k_trans.nii.gz',
                                          average=k_trans_avg,
                                          label_map=ktrans_label_map)
        chisq_label_map = LabelMap(filename='k_trans_overlay.nii.gz',
                                   color_table='jet.txt')
        chisq = Modeling.ParameterResult(filename='chi_sq.nii.gz',
                                         average=chi_sq_avg,
                                         label_map=chisq_label_map)
        result = dict(ktrans=ktrans, chisq=chisq)
        session.modeling = Modeling(protocol=tofts, source=t1,
                                    resource='pk_h7Jtl', result=result)

    It is then the responsibility of an imaging web app REST read client
    to interpret the modeling result dictionary items and display them
    appropriately.
    
    .. reST substitutions:
    .. include:: <isogrk3.txt>
    .. |H2O| replace:: H\ :sub:`2`\ O
    .. |Ktrans| replace:: K\ :sup:`trans`
    .. |ve| replace:: v\ :sub:`e`
    .. |taui| replace:: |tau|\ :sub:`i`
    .. |chisq| replace:: |chi|\ :sup:`2`

    .. _OHSU QIN modeling workflow: http://qipipe.readthedocs.org/en/latest/api/pipeline.html#modeling
    .. _Tofts standard: http://onlinelibrary.wiley.com/doi/10.1002/(SICI)1522-2586(199909)10:3%3C223::AID-JMRI2%3E3.0.CO;2-S/abstract
    .. _shutter speed: http://www.ncbi.nlm.nih.gov/pmc/articles/PMC2582583
    """

    def __str__(self):
        return "Modeling %s" % self.resource


class SessionDetail(mongoengine.Document):
    """The MR session detailed content."""

    meta = dict(collection='qiprofile_session_detail')

    scans = fields.ListField(field=mongoengine.EmbeddedDocumentField(Scan))
    """The list of scans."""



"""
The qiprofile clinical Mongodb data model.
"""

import re
import mongoengine
from mongoengine import (fields, ValidationError)
from .. import choices
from .encounter import Encounter

POS_NEG_CHOICES = [(True, 'Positive'), (False, 'Negative')]
"""The Boolean choices for Positive/Negative display values."""

YES_NO_CHOICES = [(True, 'Yes'), (False, 'No')]
"""The Boolean choices for Yes/No display values."""


class Treatment(mongoengine.EmbeddedDocument):
    """
    The patient therapy, e.g. adjuvant. Treatment is one of
    the :const:`Treatment.TYPE_CHOICES` types, and occurs over
    a period of time. The treatment consists of dosages, which
    may be pharmocological or radiological.
    """

    TYPE_CHOICES = ('Neoadjuvant', 'Primary', 'Adjuvant')

    treatment_type = fields.StringField(choices=TYPE_CHOICES)

    start_date = fields.DateTimeField(required=True)

    end_date = fields.DateTimeField(required=True)

    dosages = fields.ListField(
        field=mongoengine.EmbeddedDocumentField('Dosage')
    )


class Dosage(mongoengine.EmbeddedDocument):
    """The agent dosage."""

    agent = fields.EmbeddedDocumentField('Agent', required=True)
    """The administered Drug or Radiation."""

    amount = fields.FloatField(required=True)
    """
    The cumulative amount of the agent administered over the
    course of the duration, normalized by weight.
    
    For chemotherapy, the field unit is milligrams per kilogram
    (mg/kg).
    
    For radiotherapy, the field unit is Greys per kilogram (Gy/kg).
    
    Radiation fractions and daily chemotherapy dosages are not tracked.
    """

    start_date = fields.DateTimeField()
    """The first date the agent is administered."""

    duration = fields.IntField()
    """
    The span in days during which the agent is administered.
    """


class Agent(mongoengine.EmbeddedDocument):
    """A treatment agent, e.g. drug or radiation."""

    meta = dict(allow_inheritance=True)


class Drug(Agent):

    name = fields.StringField(required=True)
    """The official listed drug name."""


class Radiation(Agent):

    BEAM_TYPES = ['photon', 'proton', 'electron', 'neutron', 'carbon']
    """
    The radiation beam type controlled values.
    """

    beam_type = fields.StringField(choices=BEAM_TYPES)


class OtherAgent(Agent):

    name = fields.StringField(required=True)


class Biopsy(Encounter):
    """Non-therapeutic tissue extraction resulting in a pathology report."""

    pathology = fields.EmbeddedDocumentField('Pathology', required=True)


class Surgery(Encounter):
    """Therapeutic tissue extraction which usually results in a pathology report."""

    meta = dict(allow_inheritance=True)

    pathology = fields.EmbeddedDocumentField('Pathology')


class Assessment(Encounter):
    """Generic collection of outcomes."""

    evaluation = fields.EmbeddedDocumentField('GenericEvaluation')


class BreastSurgery(Surgery):
    """Breast tumor extraction."""

    TYPE_CHOICES = ('Total Mastectomy', 'Partial Mastectomy', 'Lumpectomy')
    """The surgery type controlled values."""

    surgery_type = fields.StringField(choices=TYPE_CHOICES)


class Evaluation(mongoengine.EmbeddedDocument):
    """The patient evaluation holds outcomes."""

    meta = dict(allow_inheritance=True)


class GenericEvaluation(Evaluation):
    """An unconstrained set of outcomes."""

    outcomes = fields.ListField(fields.EmbeddedDocumentField('Outcome'))


class Pathology(Evaluation):
    """The patient pathology summary."""

    meta = dict(allow_inheritance=True)

    tnm = fields.EmbeddedDocumentField('TNM')


class BreastGeneticExpression(mongoengine.EmbeddedDocument):
    """The breast patient genetic expression results."""

    HER2_NEU_IHC_CHOICES = [(0, '0'), (1, '1+'), (2, '2+'), (3, '3+')]
    """The HER2 NEU IHC choices are displayed as 0, 1+, 2+, 3+."""

    class KI67Field(fields.IntField):
        def validate(self, value, clean=True):
            return value >= 0 and value <= 100

    her2_neu_ihc = fields.IntField(choices=HER2_NEU_IHC_CHOICES)

    her2_neu_fish = fields.BooleanField(choices=POS_NEG_CHOICES)

    ki67 = KI67Field()

    normalized_assay = fields.EmbeddedDocumentField('BreastNormalizedAssay')


class BreastPathology(Pathology):
    """The QIN breast patient pathology summary."""

    hormone_receptors = fields.ListField(
        field=mongoengine.EmbeddedDocumentField('HormoneReceptorStatus')
    )

    genetic_expression = fields.EmbeddedDocumentField('BreastGeneticExpression')


class BreastNormalizedAssayField(fields.IntField):
    """The normalized Breast genomics result in the inclusive range [0, 15]."""

    def validate(self, value, clean=True):
        return value > 0 and value <= 15


class BreastNormalizedAssay(mongoengine.EmbeddedDocument):
    """The Breast genomics panel normalized to reference genes."""

    class HER2(mongoengine.EmbeddedDocument):
        grb7 = BreastNormalizedAssayField()
        her2 = BreastNormalizedAssayField()

    class Estrogen(mongoengine.EmbeddedDocument):
        er = BreastNormalizedAssayField()
        pgr = BreastNormalizedAssayField()
        bcl2 = BreastNormalizedAssayField()
        scube2 = BreastNormalizedAssayField()

    class Proliferation(mongoengine.EmbeddedDocument):
        ki67 = BreastNormalizedAssayField()
        stk15 = BreastNormalizedAssayField()
        survivin = BreastNormalizedAssayField()
        ccnb1 = BreastNormalizedAssayField()
        mybl2 = BreastNormalizedAssayField()

    class Invasion(mongoengine.EmbeddedDocument):
        mmp11 = BreastNormalizedAssayField()
        ctsl2 = BreastNormalizedAssayField()

    gstm1 = BreastNormalizedAssayField()

    cd68 = BreastNormalizedAssayField()

    bag1 = BreastNormalizedAssayField()

    her2 = fields.EmbeddedDocumentField(HER2)

    estrogen = fields.EmbeddedDocumentField(Estrogen)

    proliferation = fields.EmbeddedDocumentField(Proliferation)

    invasion = fields.EmbeddedDocumentField(Invasion)


class SarcomaPathology(Pathology):
    """The QIN sarcoma patient pathology summary."""

    HISTOLOGY_CHOICES = ('Carcinosarcoma', 'Cerebellar', 'Chondrosarcoma',
                         'Clear Cell', 'Dermatofibrosarcoma', 'Fibrosarcoma',
                         'Leiomyosarcoma', 'Liposarcoma', 'MFH', 'MPNST',
                         'Osteosarcoma', 'Rhabdomyosarcoma', 'Synovial', 'Other')
    """The histololgy controlled values."""

    location = fields.StringField()

    histology = fields.StringField(choices=HISTOLOGY_CHOICES)

    necrosis_percent = fields.EmbeddedDocumentField('NecrosisPercent')


## Clinical metrics ##

class Outcome(mongoengine.EmbeddedDocument):
    """The patient clinical outcome."""

    meta = dict(allow_inheritance=True)


class TNM(Outcome):
    """
    The TNM tumor staging. The TNM fields are as follows:

      * size - primary tumor size (T)

      * lymph_status - regional lymph nodes (N)

      * metastasis - distant metastasis (M)

      * grade - tumor grade (G)

      * serum_tumor_markers (S)

      * resection_boundaries (R)

      * lymphatic_vessel_invasion (L)

      * vein_invasion (V)

    The size is an aggregate Size field.
    See http://www.cancer.gov/cancertopics/factsheet/detection/staging for
    an overview. See http://en.wikipedia.org/wiki/TNM_staging_system and
    http://cancerstaging.blogspot.com/ for the value definition.

    :Note: The size and lymph_status choices can be further constrained by
        tumor type. Since :class:`TNM` is a generic class, these constraints
        are not enforced in this TNM class. Rather, the REST client is
        responsible for enforcing additional choice constraints. The
        :meth:`TNM.lymph_status_choices` helper method can be used for
        tumor type specific choices. See :class:`TNM.Size`` for a discussion
        of the size constraints.
  """
    class Size(mongoengine.EmbeddedDocument):
        """
        The TNM primary tumor size field.

        :Note: The size score choices can be further constrained by tumor
            type. For example, the sarcoma tumor_size choices are 0, 1 or 2
            and suffix choices are ``a`` or ``b``. See :class:`TNM` for a
            discussion of choice constraints. The :meth:`TNM.Size.tumor_size_choices`
            and :meth:`TNM.Size.suffix_choices` helper methods can be used for
            tumor type specific choices.
        """

        PREFIXES = ['c', 'p', 'y', 'r', 'a', 'u']

        SUFFIXES = ['a', 'b', 'c']

        SUFFIX_CHOICES = dict(
            Any=['a', 'b', 'c'],
            Sarcoma=['a', 'b']
        )

        TUMOR_SIZE_CHOICES = dict(
            Any=range(0, 5),
            Sarcoma=range(0, 3)
        )

        SIZE_PAT = """
            ^(
             (?P<prefix>c|p|y|r|a|u)?   # The prefix modifier
             T)?                        # The size designator
            (x |                        # Size cannot be evaluated
             (?P<in_situ>is) |          # Carcinoma in situ
             ((?P<tumor_size>0|1|2|3|4) # The size
              (?P<suffix>a|b|c)?        # The suffix modifier
             )
            )$
        """
        """
        The tumor size pattern.
        
        Examples:
        * ``T3``
        * ``pT2`` - pathology prefix
        * ``T3a`` - ``a``, ``b`` or ``c`` suffix modifier is allowed
        * ``3a`` - ``T`` prefix is optional for the size alone
        * ``Tx`` - tumor size cannot be evaluated
        * ``Tis`` - in situ
        """

        SIZE_REGEX = re.compile(SIZE_PAT, re.VERBOSE)
        """The :const:`SIZE_PAT` pattern regular expression."""

        @staticmethod
        def tumor_size_choices(tumor_type=None):
            """
            :param tumor_type: the optional tumor type, e.g. ``Breast``
            :return: the tumor_size choices for the given type
            """
            if tumor_type not in TNM.Size.TUMOR_SIZE_CHOICES:
                tumor_type = 'Any'

            return TNM.Size.TUMOR_SIZE_CHOICES[tumor_type]

        @staticmethod
        def suffix_choices(tumor_type=None):
            """
            :param tumor_type: the optional tumor type, e.g. ``Breast``
            :return: the suffix choices for the given type
            """
            if tumor_type not in TNM.Size.SUFFIX_CHOICES:
                tumor_type = 'Any'

            return TNM.Size.SUFFIX_CHOICES[tumor_type]

        prefix = fields.StringField(choices=PREFIXES)

        tumor_size = fields.IntField(choices=TUMOR_SIZE_CHOICES['Any'])

        class InSitu(mongoengine.EmbeddedDocument):
            INVASIVE_TYPE_CHOICES = ('ductal', 'lobular')
            """
            The advisory invasive types list. The client should constraion the invasive
            type choices to this list where possible, but allow for free-form text where
            necessary.
            """

            invasive_type = fields.StringField()

        in_situ = fields.EmbeddedDocumentField(InSitu)

        suffix = fields.StringField(choices=SUFFIX_CHOICES['Any'])

        def __str__(self):
            prefix = self.prefix or ''
            suffix = self.suffix or ''
            if self.in_situ:
                size = 'is'
            elif self.tumor_size:
                size = str(self.tumor_size)
            else:
                size = 'x'

            return "%sT%s%s" % (prefix, size, suffix)

        @classmethod
        def parse(klass, value):
            """
            Parses the given string into a new Size. The size must match
            the :const:`SIZE_REGEX` regular expression.

            :param value: the input string
            :return: the new Size object
            :raise ValidationError: it the size value string does not
                match :const:`SIZE_REGEX`
            """
            match = klass.SIZE_REGEX.match(value)
            if not match:
                raise ValidationError("TNM Size value is not supported:"
                                      " %s" % value)
            
            return klass(**match.groupdict())

        def clean(self):
            """
            Peforms document-level validation.

            :raise ValidationError: if the in_situ flag is set but there
              is a tumor_size or suffix field
            """
            if self.in_situ:
                if self.tumor_size != None:
                    raise ValidationError("TNM Size with in_situ flag set to"
                                          " True cannot have tumor_size %d" %
                                          self.tumor_size)
                if self.suffix != None:
                    raise ValidationError("TNM Size with in_situ flag set to"
                                          " True cannot have a suffix %s" %
                                          self.suffix)
            return True

    LYMPH_STATUS_CHOICES = dict(
        Any=range(0, 4),
        Sarcoma=range(0, 2)
    )

    tumor_type = fields.StringField(required=True)

    size = fields.EmbeddedDocumentField(Size)

    # TODO - make lymph status an aggregate with suffix modifiers,
    # including 'mi'.
    lymph_status = fields.IntField(choices=LYMPH_STATUS_CHOICES['Any'])

    metastasis = fields.BooleanField(choices=POS_NEG_CHOICES)

    grade = fields.EmbeddedDocumentField('Grade')

    serum_tumor_markers = fields.IntField(choices=range(0, 4))

    resection_boundaries = fields.IntField(choices=range(0, 3))

    lymphatic_vessel_invasion = fields.BooleanField()

    vein_invasion = fields.IntField(choices=range(0, 3))

    @staticmethod
    def lymph_status_choices(tumor_type=None):
        """
        :param tumor_type: the optional tumor type, e.g. ``Breast``
        :return: the lymph_status choices for the given type
        """
        if tumor_type not in TNM.LYMPH_STATUS_CHOICES:
            tumor_type = 'Any'

        return TNM.LYMPH_STATUS_CHOICES[tumor_type]


class Grade(Outcome):
    """
    The abstract tumor grade superclass, specialized for each
    tumor type.
    """

    meta = dict(allow_inheritance=True)


class ModifiedBloomRichardsonGrade(Grade):
    """
    The `Modified Bloom Richardson <http://pathology.jhu.edu/breast/grade.php>`_
    (a.k.a. Nottingham) breast tumor grade.
    """

    COMPONENT_CHOICES = range(1, 4)

    tubular_formation = fields.IntField(choices=COMPONENT_CHOICES)

    nuclear_pleomorphism = fields.IntField(choices=COMPONENT_CHOICES)

    mitotic_count = fields.IntField(choices=COMPONENT_CHOICES)


class HormoneReceptorStatus(Outcome):
    """The patient estrogen/progesterone hormone receptor status."""

    class IntensityField(fields.IntField):
        def validate(self, value, clean=True):
            return value > 0 and value <= 100

    hormone = fields.StringField(required=True)

    positive = fields.BooleanField(choices=YES_NO_CHOICES)

    quick_score = fields.IntField(choices=range(0, 9))

    intensity = IntensityField()


class FNCLCCGrade(Grade):
    """The FNCLCC sarcoma tumor grade."""

    differentiation = fields.IntField(choices=range(1, 4))

    necrosis_score = fields.IntField(choices=range(0, 3))

    mitotic_count = fields.IntField(choices=range(1, 4))


def necrosis_percent_as_score(necrosis_percent):
    """
    Calculates the necrosis score from the necrosis percent
    according to the
    `Stanford Synovial Sarcoma Guideline<http://surgpathcriteria.stanford.edu/softmisc/synovial_sarcoma/grading.html>`
    as follows:
    * If the percent is None, then None
    * Otherwise, if the percent is 0, then 0
    * Otherwise, if the percent is less than 50, then 1
    * Otherwise, 2
    
    :param necrosis_percent: the integer percent,
        :class:`NecrosisPercentValue` or  :class:`NecrosisPercentRange`
    :return: the necrosis score
    :raise ValidationError: if the percent is a range that spans 50%
    """
    if necrosis_percent == None:
        return None
    # Wrap a simple integer as a trivial range.
    if isinstance(necrosis_percent, int):
        necrosis_range = NecrosisPercentRange(
            start=NecrosisPercentRange.LowerBound(value=necrosis_percent),
            stop=NecrosisPercentRange.UpperBound(value=necrosis_percent + 1)
        )
    # Convert a value to a trivial range for convenience.
    elif isinstance(necrosis_percent, NecrosisPercentValue):
        necrosis_range = NecrosisPercentRange(
            start=NecrosisPercentRange.LowerBound(value=necrosis_percent.value),
            stop=NecrosisPercentRange.UpperBound(value=necrosis_percent.value + 1)
        )
    elif isinstance(necrosis_percent, NecrosisPercentRange):
        necrosis_range = necrosis_percent
    else:
        raise ValidationError("Necrosis percent type is not supported: %s" %
                              necrosis_percent.__class__)
    if necrosis_range.stop.value == 1:
        return 0
    elif necrosis_range.stop.value <= 50:
        return 1
    elif necrosis_range.start.value >= 50:
        return 2
    else:
        raise ValidationError("The necrosis percent score cannot be"
                              " determined from the range" % necrosis_range)


class NecrosisPercent(Outcome):
    """The necrosis percent value or range."""

    meta = dict(allow_inheritance=True)


class NecrosisPercentValue(NecrosisPercent):
    """The necrosis percent absolute value."""

    value = fields.IntField(choices=range(0, 101))


class NecrosisPercentRange(NecrosisPercent):
    """
    The necrosis percent range.
    
    :Note: it is recommended, although not required, that the percent
        range is a decile range, e.g. [20-30].
    
    :Note: A range which spans 50%, e.g. [40-60], results in a
        :meth:`necrosis_percent_as_score` ValidationError.
    """

    class Bound(mongoengine.EmbeddedDocument):
        """
        Necrosis percent upper or lower bound abstract class.
        The subclass is responsible for adding the ``inclusive``
        field.
        """

        meta = dict(allow_inheritance=True)

        value = fields.IntField(choices=range(0, 101))


    class LowerBound(Bound):
        """Necrosis percent lower bound."""

        inclusive = fields.BooleanField(default=True)


    class UpperBound(Bound):
        """Necrosis percent upper bound."""

        inclusive = fields.BooleanField(default=False)

    start = fields.EmbeddedDocumentField(LowerBound)

    stop = fields.EmbeddedDocumentField(UpperBound)
    
    def __repr__(self):
        return "%d-%d" % (self.start, self.stop)



"""
The qiprofile administrative Mongodb data model.
"""

import mongoengine
from mongoengine import fields

class User(mongoengine.Document):
    """
    The application user.
    """

    meta = dict(collection='qiprofile_user')

    email = fields.StringField()
    first_time = fields.BooleanField(default=True)

    def __str__(self):
        return self.email





import os
import re
import glob
from setuptools import (setup, find_packages)
from pip.req import parse_requirements

def version(package):
    """
    Return package version as listed in the `__init__.py` `__version__`
    variable.
    """
    init_py = open(os.path.join(package, '__init__.py')).read()
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)
    

def requires():
    with open('requirements.txt') as f:
      return f.read().splitlines()


def readme():
    with open("README.rst") as f:
        return f.read()


setup(
    name = 'qiprofile-rest-client',
    version = version('qiprofile_rest_client'),
    author = 'OHSU Knight Cancer Institute',
    author_email = 'loneyf@ohsu.edu',
    packages = find_packages(),
    include_package_data = True,
    scripts = glob.glob('bin/*'),
    url = 'http://qiprofile-rest-client.readthedocs.org/en/latest/',
    description = 'Quantitative Imaging Profile REST client',
    long_description = readme(),
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Environment :: Web Environment',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Natural Language :: English'
    ],
    install_requires = requires()
)
